from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .query import Field
    from .typing import Clause


# -- Clause


def ensure_dict(data: dict[str, Any] | Clause) -> dict[str, Any]:
    return data.to_mongo() if hasattr(data, "to_mongo") else data


# -- Base operator classes


class ClauseOperator:
    """
    Base class for clause sequence based operators.
    """

    __slots__ = ("_clauses",)

    _operator: str = None  # type: ignore[assignment]

    def __init__(self, *clauses: Clause) -> None:
        self._clauses = clauses

    def __init_subclass__(cls) -> None:
        if cls._operator is None:
            cls._operator = f"${cls.__name__.lower()}"

    @property
    def clauses(self) -> Generator[Clause, None, None]:
        for clause in self._clauses:
            yield clause

    def to_mongo(self) -> dict[str, Any]:
        return {self._operator: [c.to_mongo() for c in self._clauses]}


class KeyValueOperator:
    """
    Base class for key-value pair based operators.
    """

    __slots__ = ("_key", "_value")

    _operator: str = None  # type: ignore[assignment]

    def __init__(self, key: str | Field, value: Any) -> None:
        self._key = key if isinstance(key, str) else key.name
        self._value = value

    def __init_subclass__(cls) -> None:
        if cls._operator is None:
            cls._operator = f"${cls.__name__.lower()}"

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> Any:
        return self._value

    def to_mongo(self) -> dict[str, Any]:
        return {self._key: {self._operator: self._value}}


# -- Raw operator


class Raw:
    """
    Clause that wraps a raw, MongoDB-compatible dict.
    """

    __slots__ = ("_data",)

    def __init__(self, value: dict[str, Any]) -> None:
        self._data = value

    def to_mongo(self) -> dict[str, Any]:
        return {**self._data}  # Just a shallow copy


# -- Comparison operators


class Eq(KeyValueOperator):
    """`$eq` MongoDB operator."""

    __slots__ = ()


class DirectEq(KeyValueOperator):
    """Plain equality operator, `{key: value}`."""

    __slots__ = ()
    _operator = ""

    def to_mongo(self) -> dict[str, Any]:
        return {self._key: self._value}


class Ne(KeyValueOperator):
    """`$ne` MongoDB operator."""

    __slots__ = ()


class Gt(KeyValueOperator):
    """`$gt` MongoDB operator."""

    __slots__ = ()


class Gte(KeyValueOperator):
    """`$gte` MongoDB operator."""

    __slots__ = ()


class Lt(KeyValueOperator):
    """`$lt` MongoDB operator."""

    __slots__ = ()


class Lte(KeyValueOperator):
    """`$lte` MongoDB operator."""

    __slots__ = ()


class In(KeyValueOperator):
    """`$in` MongoDB operator."""

    __slots__ = ()


class NotIn(KeyValueOperator):
    """`$nin` MongoDB operator."""

    __slots__ = ()
    _operator = "$nin"


# -- Logical operators


class And(ClauseOperator):
    """`$and` MongoDB operator."""

    __slots__ = ()


class Not(ClauseOperator):
    """`$not` MongoDB operator."""

    __slots__ = ()


class Or(ClauseOperator):
    """`$or` MongoDB operator."""

    __slots__ = ()


class Nor(ClauseOperator):
    """`$nor` MongoDB operator."""

    __slots__ = ()


# -- Element operators


class Exists(KeyValueOperator):
    """`$exists` MongoDB operator."""

    __slots__ = ()

    def __init__(self, key: str | Field, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError("Exists field only accepts bool values.")

        super().__init__(key, value)


class Type(KeyValueOperator):
    """`$type` MongoDB operator."""

    __slots__ = ()

    def __init__(self, key: str | Field, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError("Type field only accepts string values.")

        super().__init__(key, value)


# -- Array operators


class All(KeyValueOperator):
    """`$all` MongoDB operator."""

    __slots__ = ()

    def __init__(self, key: str | Field, value: list[Any]) -> None:
        if not isinstance(value, list):
            raise ValueError("All field only accepts list values.")
        super().__init__(key, value)


class ElemMatch(KeyValueOperator):
    """`$elemMatch` MongoDB operator."""

    __slots__ = ()
    _operator = "$elemMatch"

    def __init__(self, key: str | Field, value: dict[str, Any]) -> None:
        if not isinstance(value, dict):
            raise ValueError("ElemMatch field only accepts dict values.")
        super().__init__(key, value)


class Size(KeyValueOperator):
    """`$size` MongoDB operator."""

    __slots__ = ()

    def __init__(self, key: str | Field, value: int) -> None:
        if not isinstance(value, int):
            raise ValueError("Size field only accepts int values.")
        super().__init__(key, value)



from collections.abc import Callable, Coroutine
from typing import Literal, TypeVar

from .bound_method_wrapper import BoundMethodWrapper
from .typing import ClauseOrMongoQuery

__all__ = (
    "ValidationError",
    "Validator",
    "validator",
)


class ValidationError(Exception):
    ...


TOwner = TypeVar("TOwner")
TInsertOrUpdate = TypeVar("TInsertOrUpdate")
InsertUpdateConfig = Literal["insert", "update", "insert-update"]


class Validator(
    BoundMethodWrapper[TOwner, [TInsertOrUpdate, ClauseOrMongoQuery | None], InsertUpdateConfig]
):
    """
    Validator method wrapper.

    Validator methods receive an insert or update object, and execute some - potentially
    async - operations to make sure the inserted or updated data is valid.
    """

    __slots__ = ()

    exception = ValidationError


def validator(
    config: InsertUpdateConfig = "insert-update",
) -> Callable[
    [Callable[[TOwner, TInsertOrUpdate, ClauseOrMongoQuery | None], Coroutine[None, None, None]]],
    "Validator[TOwner, TInsertOrUpdate]",
]:
    """
    Service method decorator factory that converts the decorated method into a `Validator` instance.

    Example:

    ```python
    class SVC(Service):
        @validator("update")
        def check_something(self, data: InsertData | CreateData, query: ClauseOrMongoQuery | None) -> None:
            raise ValueError("Always fail.")
    ```

    Arguments:
        config: Validatator config.
    """

    def decorator(
        func: Callable[[TOwner, TInsertOrUpdate, ClauseOrMongoQuery | None], Coroutine[None, None, None]],
        /,
    ) -> "Validator[TOwner, TInsertOrUpdate]":
        return Validator(func=func, config=config)

    return decorator



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from pydantic import BaseModel

from .operator import (
    All,
    And,
    DirectEq,
    ElemMatch,
    Eq,
    Exists,
    Gt,
    Gte,
    In,
    Lt,
    Lte,
    Ne,
    NotIn,
    Or,
    Size,
    Type,
)
from .typing import Clause

_T = TypeVar("_T", bound=BaseModel)

if TYPE_CHECKING:
    from .typing import MongoQuery


class Field:
    """
    Queryable field with support for MongoDB operators in the form of
    overridden built-in operators or MongoDB-specific methods.
    """

    __slots__ = ("_name",)

    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        """
        The name of the field as well as the name of the corresponding property
        in the MongoDB model.
        """
        return self._name

    # -- Operators

    def __lt__(self, value: Any) -> Query:
        return Query(Lt(self._name, value))

    def __le__(self, value: Any) -> Query:
        return Query(Lte(self._name, value))

    def __eq__(self, value: Any) -> Query:  # type: ignore[override]
        if isinstance(value, (list, set, tuple, dict)):
            return Query(Eq(self._name, value))

        return Query(DirectEq(self._name, value))

    def __ne__(self, value: Any) -> Query:  # type: ignore[override]
        return Query(Ne(self._name, value))

    def __gt__(self, value: Any) -> Query:
        return Query(Gt(self._name, value))

    def __ge__(self, value: Any) -> Query:
        return Query(Gte(self._name, value))

    # -- Methods

    def All(self, value: list[Any]) -> Query:
        """
        `$all` operator.

        Arguments:
            value: The operator's argument.
        """
        return Query(All(self._name, value))

    def ElemMatch(self, value: dict[str, Any]) -> Query:
        """
        `$elemMatch` operator.

        Arguments:
            value: The operator's argument.
        """
        return Query(ElemMatch(self._name, value))

    def Exists(self, value: bool) -> Query:
        """
        `$exists` operator.

        Arguments:
            value: The operator's argument.
        """
        return Query(Exists(self._name, value))

    def In(self, value: Any) -> Query:
        """
        `$in` operator.

        Arguments:
            value: The operator's argument.
        """
        return Query(In(self._name, value))

    def NotIn(self, value: Any) -> Query:
        """
        `$nin` operator.

        Arguments:
            value: The operator's argument.
        """
        return Query(NotIn(self._name, value))

    def Size(self, value: int) -> Query:
        """
        `$size` operator.

        Arguments:
            value: The operator's argument.
        """
        return Query(Size(self._name, value))

    def Type(self, value: str) -> Query:
        """
        `$type` operator.

        Arguments:
            value: The operator's argument.
        """
        return Query(Type(self._name, value))


class Query:
    """
    Query implementation.
    """

    __slots__ = ("_clause",)

    def __init__(self, clause: Clause | None = None) -> None:
        """
        Initialization.

        Arguments:
            clause: The default clause of the query.
        """
        self._clause = clause

    def __and__(self, value: Clause) -> Query:
        result = self.clone()
        clause = value._clause if isinstance(value, Query) else value
        if clause is None:
            result._clause = self._clause
        elif isinstance(self._clause, And):
            result._clause = And(*self._clause.clauses, clause)
        else:
            result._clause = clause if self._clause is None else And(self._clause, clause)

        return result

    def __or__(self, value: Clause) -> Query:
        result = self.clone()
        clause = value._clause if isinstance(value, Query) else value
        if clause is None:
            result._clause = self._clause
        elif isinstance(self._clause, Or):
            result._clause = Or(*self._clause.clauses, clause)
        else:
            result._clause = clause if self._clause is None else Or(self._clause, clause)

        return result

    def clone(self) -> Query:
        """
        Returns a clone of the query.
        """
        result = type(self)()
        self._clone_attributes(result)
        return result

    def _clone_attributes(self, item: Query) -> None:
        """
        Clones the attributes of the query.

        Arguments:
            item: The query on which clonable properties must be set.
        """
        item._clause = self._clause

    def to_mongo(self) -> MongoQuery:
        """
        Returns the `MongoQuery` representation of the query.
        """
        return {} if self._clause is None else self._clause.to_mongo()


class Queryable:
    """
    Base class for queryable models.

    It expects a Pydantic `BaseModel` class during subclassing in its `model` argument and
    adds a queryable `Field` descriptor to the subclass for each field of the Pydantic model.

    Example:

    ```python
    from pydantic import BaseModel

    class Person(BaseModel):
        name: str
        lucky_number: int

    class QPerson(Queryable, model=Person):
        __slots__ = ()
    ```

    It's recommended to use the `Q` helper to achieve the above result, so you can have
    at least ORM-level type hints and autocompletion for the created queryable class.

    ```python
    from pydantic import BaseModel

    class Person(BaseModel):
        name: str
        lucky_number: int

    QPerson = Q(Person)
    ```
    """

    __slots__ = ()

    def __init_subclass__(cls, *, model: type[BaseModel]) -> None:
        """
        Subclass initialization.

        Arguments:
            model: The Pydantic model the `Queryable` is created for.
        """
        super().__init_subclass__()
        for name, info in model.model_fields.items():
            field_name = name
            if name == "id" and info.alias == "_id":
                field_name = "_id"

            setattr(cls, name, Field(field_name))


# TODO: fix type hint when the Intersection and Not types are available.
def Q(model: type[_T]) -> type[_T]:
    """
    Creates a new `Queryable` class that can be used to construct queries for documents
    with the given Pydantic model.

    The created `Queryable` class will have a property (`Field`) for all properties
    that were declared in the given Pydantic class. For usage examples and more details,
    plase see `Queryable`.

    Arguments:
        model: The Pydantic model the `Queryable` class should be created for.
    """

    class Result(Queryable, model=model):
        __slots__ = ()

    return Result  # type: ignore[return-value]
